@@ -2,9 +2,9 @@ class String
   def irc_downcase(mapping)
     case mapping
     when :rfc1459
-      self.tr("A-Z[]\\^", "a-z{}|~")
+      self.tr("A-Z[]\\\\^", "a-z{}|~")
     when :"strict-rfc1459"
-      self.tr("A-Z[]\\", "a-z{}|")
+      self.tr("A-Z[]\\\\", "a-z{}|")
     else
       # when :ascii or unknown/nil
       self.tr("A-Z", "a-z")
@@ -16,9 +16,9 @@ class String
     when :ascii
       self.tr("a-z", "A-Z")
     when :rfc1459
-      self.tr("a-z{}|~", "A-Z[]\\^")
+      self.tr("a-z{}|~", "A-Z[]\\\\^")
     when :strict-rfc1459
-      self.tr("a-z{}|", "A-Z[]\\")
+      self.tr("a-z{}|", "A-Z[]\\\\")
     end
   end
 end
